from dataclasses import dataclass
from heapq import heappop, heappush
from itertools import count
from typing import Callable, Optional

from ride_pfa.path_finding.pfa import Path, PathFindingCls

__all__ = [
    'Dijkstra',
    'BiDijkstra',
    'AStar'
]


@dataclass
class Dijkstra(PathFindingCls):

    def find_path_cls(self,
                      start: int,
                      end: int,
                      cms: Optional[set[int]] = None) -> Path:
        if start == end:
            return 0.0, [start]
        weight: str = self.weight
        cluster: str = self.cluster
        graph = self.g
        adjacency = graph._adj
        nodes = graph.nodes()
        c = count()
        push = heappush
        pop = heappop
        dist = {}
        pred = {}
        fringe = []
        push(fringe, (0.0, next(c), 0, start, None))
        while fringe:
            (d, _, n, v, p) = pop(fringe)
            if v in dist:
                continue
            dist[v] = (d, n)
            pred[v] = p
            if v == end:
                break
            for u, e in adjacency[v].items():
                if cms and nodes[u][cluster] not in cms:
                    continue
                vu_dist = d + e[weight]
                if u not in dist:
                    push(fringe, (vu_dist, next(c), n + 1, u, v))
        d, n = dist[end]
        n += 1
        path = [None] * n
        i = n - 1
        e = end
        while i >= 0:
            path[i] = e
            i -= 1
            e = pred[e]
        return d, path


@dataclass
class BiDijkstra(PathFindingCls):
    def find_path_cls(self,
                      start: int,
                      end: int,
                      cms: Optional[set[int]] = None) -> Path:
        if start == end:
            return 0, [start]

        weight: str = self.weight
        cluster: str = self.cluster
        push = heappush
        pop = heappop
        dist = ({start: (0, 0, None)}, {end: (0, 0, None)})
        fringe = ([], [])
        c = count()

        graph = self.g

        adjacency = graph._adj
        nodes = graph.nodes()

        push(fringe[0], (0, next(c), 0, start))
        push(fringe[1], (0, next(c), 0, end))

        union_node = None
        union_dst = float('inf')
        while fringe[0] and fringe[1]:
            (d1, _, n1, v1) = pop(fringe[0])
            (d2, _, n2, v2) = pop(fringe[1])
            for u, e in adjacency[v1].items():
                if cms and nodes[u][cluster] not in cms:
                    continue

                vu_dist = d1 + e[weight]
                if u not in dist[0] or dist[0][u][0] > vu_dist:
                    dist[0][u] = (vu_dist, n1 + 1, v1)
                    push(fringe[0], (vu_dist, next(c), n1 + 1, u))
                if u in dist[1]:
                    dd = dist[1][u][0] + dist[0][u][0]
                    if dd < union_dst:
                        union_dst = dd
                        union_node = u
            for u, e in adjacency[v2].items():
                if cms and nodes[u][cluster] not in cms:
                    continue

                vu_dist = d2 + e[weight]
                if u not in dist[1] or dist[1][u][0] > vu_dist:
                    dist[1][u] = (vu_dist, n2 + 1, v2)
                    push(fringe[1], (vu_dist, next(c), n2 + 1, u))
                if u in dist[0]:
                    dd = dist[0][u][0] + dist[1][u][0]
                    if dd < union_dst:
                        union_dst = dd
                        union_node = u
            if d1 + d2 > union_dst:
                break

        d1, n1, _ = dist[0][union_node]
        d2, n2, _ = dist[1][union_node]
        path = [0] * (n1 + n2 + 1)
        e = union_node
        i = n1
        while dist[0][e][2] is not None:
            path[i] = e
            i -= 1
            e = dist[0][e][2]
        path[0] = e

        e = union_node
        i = n1
        while dist[1][e][2] is not None:
            path[i] = e
            i += 1
            e = dist[1][e][2]
        path[-1] = e
        return union_dst, path


def zero(u, v):
    return 0


@dataclass
class AStar(PathFindingCls):
    h: Callable[[int, int], float] = zero

    def find_path_cls(self,
                      start: int,
                      end: int,
                      cms: Optional[set[int]] = None) -> Path:
        if start == end:
            return 0.0, [start]

        weight: str = self.weight
        cluster: str = self.cluster

        h = self.h
        graph = self.g
        adjacency = graph._adj
        nodes = graph.nodes()
        c = count()
        push = heappush
        pop = heappop
        fringe = []

        push(fringe, (0.0, next(c), 0.0, start))
        dist = {start: (0, None)}
        while fringe:
            (_, _, d, v) = pop(fringe)
            if v == end:
                break
            for u, e in adjacency[v].items():
                if cms and nodes[u][cluster] not in cms:
                    continue
                vu_dist = d + e[weight]
                if u not in dist or vu_dist < dist[u][0]:
                    dist[u] = (vu_dist, v)
                    push(fringe, (vu_dist + h(u, end), next(c), vu_dist, u))

        path = []
        e = end
        while e is not None:
            path.append(e)
            e = dist[e][1]
        return d, path[::-1]



from dataclasses import dataclass

import networkx as nx
from tqdm.auto import tqdm

from ride_pfa import utils
from ride_pfa.clustering import AbstractCommunityResolver, Community
from ride_pfa.path_finding.advenced_pfa import PathFindingAdvanced


__all__ = [
    "CentroidGraph",
    "CentroidGraphBuilder"
]


@dataclass
class CentroidGraph:
    g: nx.Graph
    cls2n: dict[int, set[int]]
    cls2c: dict[int, int]
    cls2hubs: dict[int, set[int]]
    cms: Community


@dataclass
class CentroidGraphBuilder:
    log: bool = False,
    name: str = 'cluster'
    weight: str = 'length'

    def build(self, g: nx.Graph, cms: AbstractCommunityResolver | Community) -> CentroidGraph:
        if isinstance(cms, AbstractCommunityResolver):
            cms = cms.resolve(g)
        cls2n = get_cluster_adjacency_matrix(g, name=self.name)
        g1, cls2c = build_center_graph(
            graph=g,
            communities=cms,
            cls2n=cls2n,

            log=self.log,
            name=self.name,
            weight=self.weight
        )
        cg = CentroidGraph(
            g=g1,
            cls2n=cls2n,
            cls2c=cls2c,
            cls2hubs=get_cls2hubs(g, name=self.name),
            cms=cms
        )
        return cg

    def build_with_time(self, g: nx.Graph, cms: AbstractCommunityResolver | Community, iterations=1) -> (
            tuple)[float, CentroidGraph]:
        return utils.get_execution_time(lambda: self.build(g, cms), iterations=iterations)


# cluster to neighboring clusters
def get_cluster_adjacency_matrix(graph: nx.Graph, name='cluster') -> dict[int: set[int]]:
    _cls2n = {}
    for u, du in graph.nodes(data=True):
        for v in graph[u]:
            dv = graph.nodes()[v]
            if dv[name] == du[name]:
                continue
            c1 = dv[name]
            c2 = du[name]
            if not (c1 in _cls2n):
                _cls2n[c1] = set()
            if not (c2 in _cls2n):
                _cls2n[c2] = set()
            _cls2n[c1].add(c2)
            _cls2n[c2].add(c1)
    return _cls2n


# cluster to point that are connected with neighboring clusters
def get_cls2hubs(graph: nx.Graph, name='cluster') -> dict[int: set[int]]:
    _cls2hubs = {}
    for u, du in graph.nodes(data=True):
        for v in graph[u]:
            dv = graph.nodes()[v]
            c1 = du[name]
            c2 = dv[name]
            if c1 == c2:
                continue
            if not (c1 in _cls2hubs):
                _cls2hubs[c1] = set()
            if not (c2 in _cls2hubs):
                _cls2hubs[c2] = set()
            _cls2hubs[c1].add(u)
            _cls2hubs[c2].add(v)
    return _cls2hubs


# build_center_graph
def build_center_graph(
        graph: nx.Graph,
        communities: Community,
        cls2n: dict[int: set[int]],
        log: bool = False,
        name: str = 'cluster',
        weight: str = 'length'
) -> tuple[nx.Graph, dict[int, int]]:
    x_graph = nx.Graph()
    cls2center = {}
    _iter = tqdm(enumerate(communities), total=len(communities), desc='find centroids') if log else enumerate(
        communities)
    for cls, _ in _iter:
        gc = extract_cluster_list_subgraph(graph, {cls}, communities)
        min_node = nx.barycenter(gc, weight=weight)[0]
        du = graph.nodes()[min_node]
        x_graph.add_node(graph.nodes()[min_node][name], **du)
        cls2center[graph.nodes()[min_node][name]] = min_node

    if len(x_graph.nodes) == 1:
        return x_graph, cls2center
    _iter = tqdm(x_graph.nodes(), desc='find edges') if log else x_graph.nodes()
    a_pfa: PathFindingAdvanced = PathFindingAdvanced(graph, weight=weight)
    for u in _iter:
        lengths = a_pfa.find_distance_to_all(cls2center[u], set(cls2center[v] for v in cls2n[u]))
        for v in cls2n[u]:
            x_graph.add_edge(u, v, length=lengths[cls2center[v]])
    return x_graph, cls2center


# extract subgraph by clusters
def extract_cluster_list_subgraph(graph: nx.Graph, cluster_number: list[int] | set[int], communities=None,
                                  cluster: str = 'cluster') -> nx.Graph:
    if communities:
        return graph.subgraph(_iter_cms(cluster_number, communities))
    else:
        nodes_to_keep = [node for node, data in graph.nodes(data=True) if data[cluster] in cluster_number]
    return graph.subgraph(nodes_to_keep)


def _iter_cms(cluster_number: list[int] | set[int], communities: list[set[int]] | tuple[set[int]]):
    for cls in cluster_number:
        for u in communities[cls]:
            yield u



import time
from typing import Any, Callable

import numpy as np

__all__ = [
    'get_optimal_clusters_number',
    'get_execution_time'
]


def get_optimal_clusters_number(nodes: int) -> int:
    alpha = 8.09 * (nodes ** (-0.48)) * (1 - 19.4 / (4.8 * np.log(nodes) + 8.8)) * nodes
    return int(alpha)


def get_execution_time(func: Callable, iterations=2, *args, **kwargs) -> tuple[float, Any]:
    result = None
    start = time.time()
    for _ in range(iterations):
        result = func(*args, **kwargs)
    end = time.time()
    return (end - start) / iterations, result



from abc import ABC, abstractmethod
from dataclasses import dataclass
from heapq import heappop, heappush
from itertools import count
from multiprocessing import Pool
from typing import Callable

import networkx as nx
import numpy as np
from tqdm.auto import tqdm

from ride_pfa.clustering import AbstractCommunityResolver, Community
from ride_pfa.path_finding.dijkstra_pfa import AStar
from ride_pfa.path_finding.pfa import PathFinding

__all__ = [
    'HeuristicBuilder',
    'MinClusterDistanceBuilder'
]


class HeuristicBuilder(ABC):
    @abstractmethod
    def build_astar(self, g: nx.Graph, cms: AbstractCommunityResolver | Community) -> PathFinding:
        pass


@dataclass
class MinClusterDistanceCallable(Callable[[int, int], float]):
    nodes: list[dict]
    d_cluster: np.ndarray
    cluster: str

    def __call__(self, u: int, v: int) -> float:
        nodes = self.nodes
        d_clusters = self.d_cluster
        n1, n2 = nodes[u], nodes[v]
        c11 = n1[self.cluster]
        c12 = n2[self.cluster]
        return d_clusters[c11, c12]


@dataclass
class MinClusterDistanceBuilder(HeuristicBuilder):
    workers: int = 4
    cluster = 'cluster'

    def build_astar(self, g: nx.Graph, cms: AbstractCommunityResolver | Community) -> PathFinding:
        if isinstance(cms, AbstractCommunityResolver):
            cms = cms.resolve(g)
        w = self.workers
        cms_points = list(range(len(cms)))
        data = [(cms_points[i::w], self.cluster, cms, g) for i in range(w)]

        if self.workers == 1:
            d_clusters = calc(data[0])
        else:
            with Pool(w) as p:
                d_clusters = sum(tqdm(p.imap_unordered(calc, data), total=len(data)))

        nodes = g.nodes()
        return AStar(g, h=MinClusterDistanceCallable(nodes, d_clusters, cluster=self.cluster))


def dijkstra_pfa_min_dst(graph: nx.Graph,
                         start: set[int],
                         weight: str = 'length'
                         ) -> \
        dict[float]:
    adjacency = graph._adj
    c = count()
    push = heappush
    pop = heappop
    dist = {}
    fringe = []
    for s in start:
        push(fringe, (0.0, next(c), s))
    while fringe:
        (d, _, v) = pop(fringe)
        if v in dist:
            continue
        dist[v] = d
        for u, e in adjacency[v].items():
            vu_dist = d + e[weight]
            if u not in dist:
                push(fringe, (vu_dist, next(c), u))
    return dist


def calc(data):
    points, name, cms, g = data
    d_clusters = np.zeros((len(cms), len(cms)))

    for u in points:
        ll = dijkstra_pfa_min_dst(g, cms[u])
        q = {}
        for v, d in ll.items():
            if g.nodes()[v][name] in q:
                q[g.nodes()[v][name]] = min(q[g.nodes()[v][name]], d)
            else:
                q[g.nodes()[v][name]] = d
        for v in range(len(cms)):
            if v in q:
                d_clusters[u, v] = q[v]
    return d_clusters



import random
from typing import Optional

import folium
import networkx as nx

from ride_pfa.clustering import Community

__all__ = [
    'draw_graph',
    'draw_path'
]


def draw_graph(
        graph: nx.Graph,
        cms: Optional[Community] = None,
        m: Optional[folium.Map] = None) -> folium.Map:
    if not m:
        x, y = 0, 0
        for _, d in graph.nodes(data=True):
            x, y = d['x'], d['y']
            break
        m = folium.Map(location=[y, x], zoom_start=12)
    if cms is not None:
        colors = generate_colors(len(cms))
        for i, c in cms:
            for u in c:
                d = graph.nodes()[u]
                folium.Circle(
                    [d['y'], d['x']],
                    popup=str(u),
                    fill=True,
                    color=colors[i]
                ).add_to(m)
    else:
        for node in graph.nodes():
            folium.Circle(
                [graph.nodes[node]['y'], graph.nodes[node]['x']],
                popup=str(node),
                fill=True
            ).add_to(m)

    for edge in graph.edges():
        folium.PolyLine(
            [[graph.nodes[edge[0]]['y'], graph.nodes[edge[0]]['x']],
             [graph.nodes[edge[1]]['y'], graph.nodes[edge[1]]['x']]],
            color='blue',
            weight=2
        ).add_to(m)

    return m


def draw_path(graph: nx.Graph, path: list[int],
              m: Optional[folium.Map],
              color: str = 'red',
              weight: float = 5) -> folium.Map:
    x = [graph.nodes[node]['x'] for node in path]
    y = [graph.nodes[node]['y'] for node in path]
    if not m:
        m = folium.Map(location=[y[0], x[0]], zoom_start=12)

    for i in range(len(path) - 1):
        folium.PolyLine([[y[i], x[i]], [y[i + 1], x[i + 1]]], color=color, weight=weight).add_to(m)

    return m


def random_color_hex():
    return "#{:06x}".format(random.randint(0, 0xFFFFFF))


def generate_colors(num_colors):
    return [random_color_hex() for _ in range(num_colors)]



from dataclasses import dataclass
from heapq import heappop, heappush
from itertools import count

import networkx as nx

__all__ = [
    'PathFindingAdvanced'
]


@dataclass
class PathFindingAdvanced:
    g: nx.Graph
    weight: str = 'length'

    def find_distance_to_all(self,
                             start: int,
                             ends: set[int]) -> dict[int, float]:

        weight: str = self.weight
        graph = self.g
        adjacency = graph._adj
        c = count()
        push = heappush
        pop = heappop
        dist = {}
        visited = set()
        fringe = []
        push(fringe, (0.0, next(c), start))
        while fringe:
            (d, _, v) = pop(fringe)
            if v in dist:
                continue
            dist[v] = d
            if v in ends:
                visited.add(v)
            if len(visited) == len(ends):
                break
            for u, e in adjacency[v].items():
                vu_dist = d + e[weight]
                if u not in dist:
                    push(fringe, (vu_dist, next(c), u))
        return {k: dist[k] for k in ends}



import os
import pickle
from typing import Optional

import networkx as nx
import osmnx as ox

__all__ = [
    "get_graph",
    "osm_cities_example"
]

osm_cities_example = {
    'Barcelona': 'R347950',
    'Paris': 'R71525',
    'Prague': 'R435514',
    'Moscow': 'R2555133',
    'Singapore': 'R17140517',
    'Berlin': 'R62422',
    'Rome': 'R41485',
    'Rio': 'R2697338',
    'Delhi': 'R1942586'
}


# load graph
def get_graph(city_id: str = 'R2555133', cache_path: Optional[str] = None) -> nx.Graph:
    id_graph = city_id
    name = f'{id_graph}.pickle'
    if cache_path is not None:
        path = os.path.join(cache_path, name)
        if os.path.exists(path):
            with open(path, 'rb') as fp:
                g: nx.Graph = pickle.load(fp)
                fp.close()
        else:
            g = _get_gr(id_graph)
            with open(path, 'wb') as fp:
                pickle.dump(g, fp)
                fp.close()
    else:
        g = _get_gr(id_graph)
    assert g is not None
    g.remove_edges_from(nx.selfloop_edges(g))
    return g


def _get_gr(city_id):
    gdf = ox.geocode_to_gdf(city_id, by_osmid=True)
    polygon_boundary = gdf.unary_union
    graph = ox.graph_from_polygon(polygon_boundary,
                                  network_type='drive',
                                  simplify=True)
    H = nx.Graph()
    # add edges in new graph. Copy only weights
    for u, d in graph.nodes(data=True):
        H.add_node(u, x=d['x'], y=d['y'])
    for u, v, d in graph.edges(data=True):
        if u == v:
            continue
        H.add_edge(u, v, length=d['length'])
    del city_id, gdf, polygon_boundary, graph
    return H



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import NewType, Optional

import networkx as nx

__all__ = [
    "Path",
    "PathMatrix",
    "PathFinding",
    "PathFindingCls"
]

Path = NewType('Path', tuple[float, list[int]])
PathMatrix = NewType('Path', dict[int, tuple[float, int]])


@dataclass
class PathFinding(ABC):
    g: nx.Graph
    weight: str = 'length'

    @abstractmethod
    def find_path(self, start: int, end: int, ) -> Path:
        pass


@dataclass
class PathFindingCls(PathFinding):
    cluster: str = 'cluster'

    def find_path(self, start: int, end: int) -> Path:
        return self.find_path_cls(start=start, end=end, cms=None)

    @abstractmethod
    def find_path_cls(self,
                      start: int,
                      end: int,
                      cms: Optional[set[int]] = None) -> Path:
        pass



import logging as log
from abc import ABC, abstractmethod
from dataclasses import dataclass
from heapq import heappop, heappush
from itertools import count as count
from typing import NewType, Union

import networkx as nx
from tqdm.auto import trange

__all__ = [
    'Community',
    'AbstractCommunityResolver',
    'LouvainCommunityResolver',
    'LouvainKMeansCommunityResolver'
]

Community = NewType('Community', Union[list[set[int]], tuple[set[int]]])


@dataclass
class AbstractCommunityResolver(ABC):
    weight: str = 'length'
    cluster: str = 'cluster'
    write_community: bool = True

    def resolve(self, g: nx.Graph) -> Community:
        cms: Community = self.do_resolve(g=g)
        if self.write_community:
            return validate_cms(g, cms, cluster_name=self.cluster)
        else:
            return cms

    @abstractmethod
    def do_resolve(self, g: nx.Graph) -> Community:
        pass


@dataclass
class LouvainCommunityResolver(AbstractCommunityResolver):
    resolution: float = 1
    seed: float = 1534

    def do_resolve(self, g: nx.Graph) -> Community:
        communities = nx.community.louvain_communities(g,
                                                       seed=self.seed,
                                                       weight=self.weight,
                                                       resolution=self.resolution)
        return communities


@dataclass
class LouvainKMeansCommunityResolver(LouvainCommunityResolver):
    max_iteration: int = 20
    print_log: bool = False

    def do_resolve(self, g: nx.Graph) -> Community:
        communities = super().resolve(g)
        return self.do_resolve_kmeans(g, communities)

    def do_resolve_kmeans(self, g: nx.Graph, communities: Community) -> Community:
        if self.print_log:
            log.info(f'communities: {len(communities)}')
        _iter = trange(self.max_iteration) if self.print_log else range(self.max_iteration)
        do = True
        for _ in _iter:
            if not do:
                continue
            centers = []
            for i, cls in enumerate(communities):
                gc = g.subgraph(communities[i])
                center = nx.barycenter(gc, weight=self.weight)[0]
                centers.append(center)

            node2cls = k_means(g, centers, weight=self.weight)
            do = False
            for u, i in node2cls.items():
                if u not in communities[i]:
                    do = True
                    break
            if not do:
                continue

            communities = [set() for _ in range(len(centers))]
            for u, c in node2cls.items():
                communities[c].add(u)
            communities = validate_cms(g, communities, cluster_name=self.cluster)
        return communities


def k_means(graph: nx.Graph,
            starts: list[int],
            weight: str) -> dict[int, int]:
    adjacency = graph._adj
    c = count()
    push = heappush
    pop = heappop
    dist = {}
    fringe = []
    node2cms = {
        s: i for i, s in enumerate(starts)
    }
    for start in starts:
        push(fringe, (0.0, next(c), 0, start, start))
    while fringe:
        (d, _, n, v, p) = pop(fringe)
        if v in dist:
            continue
        node2cms[v] = node2cms[p]
        dist[v] = (d, n)
        for u, e in adjacency[v].items():
            vu_dist = d + e[weight]
            if u not in dist:
                push(fringe, (vu_dist, next(c), n + 1, u, v))
    return node2cms


def validate_cms(
        graph: nx.Graph,
        communities: Community,
        cluster_name: str) -> Community:
    cls = []
    for i, c in enumerate(communities):
        for n in nx.connected_components(graph.subgraph(c)):
            cls.append(n)
    for i, ids in enumerate(cls):
        for j in ids:
            graph.nodes()[j][cluster_name] = i
    return cls



from networkx import Graph

from ride_pfa.path_finding.pfa import PathFinding, Path, PathFindingCls

__all__ = [
    'ExtractionPfa'
]


class ExtractionPfa(PathFinding):

    def __init__(self, g: Graph, upper: PathFinding,
                 down: PathFindingCls,
                 weight: str = 'weight',
                 cluster: str = 'cluster'):
        super().__init__(g=g, weight=weight)
        self.cluster: str = cluster
        self.upper: PathFinding = upper
        self.down: PathFindingCls = down

    def find_path(self, start: int, end: int) -> Path:
        cluster = self.cluster
        nodes = self.g.nodes()
        c1, c2 = nodes[start][cluster], nodes[end][cluster]
        _, path = self.upper.find_path(start=c1, end=c2)
        return self.down.find_path_cls(start=start, end=end, cms=set(p for p in path))
